from typing import Any, Callable

from calculator.constants import SUPPORTED_INPUT_DATATYPES
from calculator.exceptions import (
    DivisionByZeroError,
    InvalidInputError,
    NegativeValueRootError,
)

Result = float
InputValue = float | int


class Calculator:
    """Calculator.

    Attributes:
        _result: A float value of Zero, default calculation value.
    """

    def __init__(self):
        self._result: Result = 0.0

    @property
    def result(self) -> Result:
        """Current calculation result, default = 0.0."""
        return self._result

    def reset(self) -> None:
        """Reset calculation result to Zero."""
        self._result = 0.0

    @staticmethod
    def _check_input_datatype(f: Callable) -> Any:
        """Function wrapper. Detects illegal argument input.

        Raises:
            InvalidInputError: If input number is not float or int datatype.
        """

        def wrapper(_, number):
            if type(number) not in SUPPORTED_INPUT_DATATYPES:
                raise InvalidInputError(number)
            res = f(_, number)
            return res

        return wrapper

    @_check_input_datatype
    def add(self, number: InputValue) -> Result:
        """Add requested number to calculation result.

        Args:
            number: A float or integer datatype value.

        Returns:
            Calculated result.
        """
        self._result += number
        return self._result

    @_check_input_datatype
    def subtract(self, number: InputValue) -> Result:
        """Subtract requested value from calculation number.

        Args:
            number: A float or integer datatype value.

        Returns:
            Calculated result.
        """
        self._result -= number
        return self._result

    @_check_input_datatype
    def multiply(self, number: InputValue) -> Result:
        """Multiply calculation result by requested number."""
        self._result *= number
        return self._result

    @_check_input_datatype
    def divide(self, number: InputValue) -> Result:
        """Divide calculation result by requested number.

        Args:
            number: A float or integer datatype value.

        Returns:
            Calculated result.

        Raises:
            DivisionByZeroError: If value is Zero.
        """
        if number == 0:
            raise DivisionByZeroError()

        self._result /= number
        return self._result

    @_check_input_datatype
    def root(self, number: InputValue) -> Result:
        """Take (number) root of calculated result.

        Args:
            number: A float or integer value.

        Returns:
            Calculated result.

        Raises:
            NegativeValueRootError: If current
                calculated result is negative.
        """
        if self._result < 0:
            raise NegativeValueRootError(self._result)

        self._result = self._result ** (1 / number)
        return self._result

    def __repr__(self):
        return f"{self._result}"

    def __str__(self):
        return f"{self._result}"
